@@ -12,7 +12,7 @@
 -- See the License for the specific language governing permissions and
 -- limitations under the License.
 --
--- Copyright (C) 2015-present, TBOOX Open Source Group.
+-- Copyright (C) 2015-present, Xmake Open Source Community.
 --
 -- @author      ruki
 -- @file        autoconf.lua
@@ -60,7 +60,11 @@ end
 function _translate_windows_bin_path(bin_path)
     if bin_path then
         local argv = os.argv(bin_path)
-        argv[1] = path.unix(argv[1]) .. ".exe"
+        argv[1] = path.unix(argv[1])
+        local path_lower = argv[1]:lower()
+        if not path_lower:endswith(".exe") and not path_lower:endswith(".cmd") and not path_lower:endswith(".bat") then
+            argv[1] = argv[1] .. ".exe"
+        end
         return os.args(argv)
     end
 end
@@ -82,18 +86,40 @@ function _memcache()
     return memcache.cache("package.tools.autoconf")
 end
 
+-- has flag ?
+function _has_flag(package, flag)
+    local flag_name = flag:gsub("-", "")
+    local has_flag = _memcache():get2(tostring(package), flag_name)
+    if has_flag == nil then
+        local result = try {function() return os.iorunv("./configure", {"--help"}, {shell = true}) end}
+        if result and result:find(flag, 1, true) then
+            has_flag = true
+        end
+        has_flag = has_flag or false
+        _memcache():set2(tostring(package), flag_name, has_flag)
+    end
+    return has_flag
+end
+
+
 -- has `--with-pic`?
 function _has_with_pic(package)
-    local has_with_pic = _memcache():get2(tostring(package), "with_pic")
-    if has_with_pic == nil then
-        local result = try {function() return os.iorunv("./configure", {"--help"}, {shell = true}) end}
-        if result and result:find("--with-pic", 1, true) then
-            has_with_pic = true
-        end
-        has_with_pic = has_with_pic or false
-        _memcache():set2(tostring(package), "with_pic", has_with_pic)
-    end
-    return has_with_pic
+    return _has_flag(package, "--with-pic")
+end
+
+-- has `--enable-debug`?
+function _has_enable_debug(package)
+    return _has_flag(package, "--enable-debug")
+end
+
+-- has `--enable-static` and `--enable-shared`?
+function _has_enable_kind(package)
+    return _has_flag(package, "--enable-static") and _has_flag(package, "--enable-shared")
+end
+
+-- has `--disable-static` and `--disable-shared`?
+function _has_disable_kind(package)
+    return _has_flag(package, "--disable-static") and _has_flag(package, "--disable-shared")
 end
 
 -- get configs
@@ -166,6 +192,21 @@ function _get_configs(package, configs)
         package:config("pic") ~= false and _has_with_pic(package) then
         table.insert(configs, "--with-pic")
     end
+    if package:is_debug() and _has_enable_debug(package) then
+        table.insert(configs, "--enable-debug")
+    end
+    if package:is_library() then
+        if _has_enable_kind(package) then
+            table.insert(configs, "--enable-shared=" .. (package:config("shared") and "yes" or "no"))
+            table.insert(configs, "--enable-static=" .. (package:config("shared") and "no" or "yes"))
+        elseif _has_disable_kind(package) then
+            if package:config("shared") then
+                table.insert(configs, "--disable-static")
+            else
+                table.insert(configs, "--disable-shared")
+            end
+        end
+    end
     return configs
 end
 
@@ -358,7 +399,7 @@ function buildenvs(package, opt)
     end
     if ldflags or shflags then
         -- autoconf does not use SHFLAGS
-        envs.LDFLAGS   = table.concat(_translate_paths(table.join(ldflags or {}, shflags)), ' ')
+        envs.LDFLAGS   = table.concat(table.reverse_unique(_translate_paths(table.join(ldflags or {}, shflags))), ' ')
     end
 
     -- cross-compilation? pass the full build environments
@@ -380,7 +421,8 @@ function buildenvs(package, opt)
                 -- force to apply shflags on macosx https://gmplib.org/manual/Known-Build-Problems
                 envs.CC = envs.CC .. " -arch " .. package:arch()
             end
-            if package:is_plat("cross") or package:has_tool("ar", "ar", "emar") then
+            -- android r27 will use llmv-ar instead of ar, https://github.com/xmake-io/xmake/issues/6206
+            if package:is_plat("cross") or package:has_tool("ar", "ar", "emar", "llvm_ar") then
                 -- only for cross-toolchain
                 envs.CXX = package:build_getenv("cxx")
                 if not envs.ARFLAGS or envs.ARFLAGS == "" then
@@ -404,10 +446,17 @@ function buildenvs(package, opt)
             name = name:gsub("clang$", "ld")
             name = name:gsub("clang%-%d+", "ld")
             name = name:gsub("gcc$", "ld")
-            name = name:gsub("gcc-%d+", "ld")
+            name = name:gsub("gcc%-%d+$", "ld")
             name = name:gsub("g%+%+$", "ld")
-            name = name:gsub("g%+%+%-%d+", "ld")
-            envs.LD = dir and path.join(dir, name) or name
+            name = name:gsub("g%+%+%-%d+$", "ld")
+            if dir and os.isfile(path.join(dir, name)) then
+                envs.LD = path.join(dir, name)
+            else
+                local ld = find_tool("ld")
+                if ld and path.filename(ld.program) == name then
+                    envs.LD = ld.program
+                end
+            end
         end
         -- we need use clang++ as cxx, autoconf will use it as linker
         -- https://github.com/xmake-io/xmake/issues/2170
@@ -421,12 +470,15 @@ function buildenvs(package, opt)
             name = name:gsub("gcc%-", "g++-")
             envs.CXX = dir and path.join(dir, name) or name
         end
-    elseif package:is_plat("windows") and not package:config("toolchains") then
+    end
+    if package:is_plat("windows") and not package:config("toolchains") then
         envs.PATH = os.getenv("PATH") -- we need to reserve PATH on msys2
         envs = os.joinenvs(envs, _get_msvc(package):runenvs())
     end
+
     if is_host("windows") then
         envs.CC       = _translate_windows_bin_path(envs.CC)
+        envs.CXX      = _translate_windows_bin_path(envs.CXX)
         envs.AS       = _translate_windows_bin_path(envs.AS)
         envs.AR       = _translate_windows_bin_path(envs.AR)
         envs.LD       = _translate_windows_bin_path(envs.LD)
