@@ -12,7 +12,7 @@
 -- See the License for the specific language governing permissions and
 -- limitations under the License.
 --
--- Copyright (C) 2015-present, TBOOX Open Source Group.
+-- Copyright (C) 2015-present, Xmake Open Source Community.
 --
 -- @author      ruki
 -- @file        vs201x.lua
@@ -40,6 +40,7 @@ import("private.action.run.runenvs")
 import("actions.config.configfiles", {alias = "generate_configfiles", rootdir = os.programdir()})
 import("private.utils.batchcmds")
 import("plugins.project.utils.target_cmds", {rootdir = os.programdir()})
+import("private.utils.target", {alias = "target_utils"})
 
 function _translate_path(dir, vcxprojdir)
     if dir == nil then
@@ -404,8 +405,10 @@ function make(outputdir, vsinfo)
                 config.set("arch", arch, {readonly = true, force = true})
 
                 -- clear all options
-                for _, opt in ipairs(project.options()) do
-                    opt:clear()
+                for _, opt in pairs(project.options()) do
+                    if not config.readonly(opt:fullname()) then
+                        opt:clear()
+                    end
                 end
 
                 -- clear cache
@@ -436,7 +439,8 @@ function make(outputdir, vsinfo)
             os.cd(project.directory())
 
             -- save targets
-            for targetname, target in table.orderpairs(project.targets()) do
+            local project_targets = target_utils.get_project_targets()
+            for targetname, target in table.orderpairs(project_targets) do
 
                 -- make target with the given mode and arch
                 targets[targetname] = targets[targetname] or {}
